/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.tqlab.sense.context;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.log4j.Logger;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.BeanFactoryUtils;
import org.springframework.beans.factory.BeanInitializationException;
import org.springframework.context.ApplicationContext;
import org.springframework.core.JdkVersion;
import org.springframework.core.OrderComparator;
import org.springframework.core.io.ClassPathResource;
import org.springframework.core.io.support.PropertiesLoaderUtils;
import org.springframework.util.ClassUtils;
import org.springframework.util.StringUtils;
import org.springframework.web.context.ContextLoader;
import org.springframework.web.servlet.DispatcherServlet;
import org.springframework.web.servlet.HandlerAdapter;
import org.springframework.web.servlet.HandlerMapping;

/**
 * Sense context Holder
 * 
 * @author John Lee
 */
public class WebApp {

	private static Logger logger = Logger.getLogger(WebApp.class);
	private final static ThreadLocal<SenseServletContext> threadLocalSenseServletContext = new ThreadLocal<SenseServletContext>();
	private final static ThreadLocal<ApplicationContext> threadLocalApplicationContext = new ThreadLocal<ApplicationContext>();

	/**
	 * 
	 * @return ApplicationContext instance
	 */
	public static ApplicationContext getApplicationContext() {
		ApplicationContext context = threadLocalApplicationContext.get();
		if (null == context) {
			context = ContextLoader.getCurrentWebApplicationContext();
		}
		return context;
	}

	/**
	 * Initial applicationContext
	 * 
	 * @param context
	 */
	public static void setApplicationContext(ApplicationContext context) {
		threadLocalApplicationContext.set(context);
	}

	/**
	 * 
	 * @return HttpServletRequest instance of current thread
	 */
	public static HttpServletRequest getHttpServletRequest() {
		SenseServletContext ssc = threadLocalSenseServletContext.get();
		if (ssc == null) {
			logger.error("SenceServletContext is null.");
			return null;
		}
		return ssc.getServletRequest();
	}

	/**
	 * 
	 * @return HttpServletResponse instance of current thread
	 */
	public static HttpServletResponse getHttpServletResponse() {
		SenseServletContext ssc = threadLocalSenseServletContext.get();
		if (ssc == null) {
			logger.error("SenceServletContext is null.");
			return null;
		}
		return ssc.getServletResponse();
	}

	/**
	 * Update SenseServletContext
	 * 
	 * @param senceServletContext
	 */
	public static void setSenseServletContext(
			SenseServletContext senceServletContext) {
		threadLocalSenseServletContext.set(senceServletContext);
	}

	/**
	 * 
	 * @return SenseServletContext instance of current thread
	 */
	public static SenseServletContext getCurrentSenceServletContext() {
		return threadLocalSenseServletContext.get();
	}

	/**
	 * Initialize the HandlerMappings used by this class.
	 * <p>
	 * If no HandlerMapping beans are defined in the BeanFactory for this
	 * namespace, we default to BeanNameUrlHandlerMapping.
	 */
	@SuppressWarnings({ "unchecked", "rawtypes" })
	public static List getHandlerMappings() {
		List handlerMappings = null;
		ApplicationContext context = getApplicationContext();

		// Find all HandlerMappings in the ApplicationContext,
		// including ancestor contexts.
		Map matchingBeans = BeanFactoryUtils.beansOfTypeIncludingAncestors(
				context, HandlerMapping.class, true, false);
		if (!matchingBeans.isEmpty()) {
			handlerMappings = new ArrayList(matchingBeans.values());
			// We keep HandlerMappings in sorted order.
			Collections.sort(handlerMappings, new OrderComparator());
		}

		// Ensure we have at least one HandlerMapping, by registering
		// a default HandlerMapping if no other mappings are found.
		if (handlerMappings == null) {
			handlerMappings = getDefaultStrategies(context,
					HandlerMapping.class);
			logger.debug("No HandlerMappings found in servlet : using default");

		}
		return handlerMappings;
	}

	/**
	 * Initialize the HandlerAdapters used by this class.
	 * <p>
	 * If no HandlerAdapter beans are defined in the BeanFactory for this
	 * namespace, we default to SimpleControllerHandlerAdapter.
	 */
	@SuppressWarnings({ "unchecked", "rawtypes" })
	public static List getHandlerAdapters() {
		List handlerAdapters = null;
		ApplicationContext context = getApplicationContext();

		// Find all HandlerAdapters in the ApplicationContext,
		// including ancestor contexts.
		Map matchingBeans = BeanFactoryUtils.beansOfTypeIncludingAncestors(
				context, HandlerAdapter.class, true, false);
		if (!matchingBeans.isEmpty()) {
			handlerAdapters = new ArrayList(matchingBeans.values());
			// We keep HandlerAdapters in sorted order.
			Collections.sort(handlerAdapters, new OrderComparator());
		}

		// Ensure we have at least some HandlerAdapters, by registering
		// default HandlerAdapters if no other adapters are found.
		if (handlerAdapters == null) {
			handlerAdapters = getDefaultStrategies(context,
					HandlerAdapter.class);
			logger.debug("No HandlerAdapters found in servlet : using default");

		}
		return handlerAdapters;
	}

	private static final Properties defaultStrategies;
	/**
	 * Name of the class path resource (relative to the DispatcherServlet class)
	 * that defines DispatcherServlet's default strategy names.
	 */
	private static final String DEFAULT_STRATEGIES_PATH = "DispatcherServlet.properties";

	static {
		// Load default strategy implementations from properties file.
		// This is currently strictly internal and not meant to be customized
		// by application developers.
		try {
			ClassPathResource resource = new ClassPathResource(
					DEFAULT_STRATEGIES_PATH, DispatcherServlet.class);
			defaultStrategies = PropertiesLoaderUtils.loadProperties(resource);
		} catch (IOException ex) {
			throw new IllegalStateException(
					"Could not load 'DispatcherServlet.properties': "
							+ ex.getMessage());
		}
	}

	/**
	 * Create a List of default strategy objects for the given strategy
	 * interface.
	 * <p>
	 * The default implementation uses the "DispatcherServlet.properties" file
	 * (in the same package as the DispatcherServlet class) to determine the
	 * class names. It instantiates the strategy objects through the context's
	 * BeanFactory.
	 * 
	 * @param context
	 *            the current WebApplicationContext
	 * @param strategyInterface
	 *            the strategy interface
	 * @return the List of corresponding strategy objects
	 * @throws BeansException
	 *             if initialization failed
	 */
	@SuppressWarnings({ "unchecked", "rawtypes" })
	private static List getDefaultStrategies(ApplicationContext context,
			Class strategyInterface) throws BeansException {
		String key = strategyInterface.getName();
		List strategies = null;
		String value = defaultStrategies.getProperty(key);
		if (value != null) {
			String[] classNames = StringUtils
					.commaDelimitedListToStringArray(value);
			strategies = new ArrayList(classNames.length);
			for (int i = 0; i < classNames.length; i++) {
				String className = classNames[i];
				if (JdkVersion.getMajorJavaVersion() < JdkVersion.JAVA_15
						&& className.indexOf("Annotation") != -1) {
					// Skip Java 5 specific strategies when running on JDK
					// 1.4...
					continue;
				}
				try {
					Class clazz = ClassUtils.forName(className,
							DispatcherServlet.class.getClassLoader());
					Object strategy = context.getAutowireCapableBeanFactory()
							.createBean(clazz);
					strategies.add(strategy);
				} catch (ClassNotFoundException ex) {
					throw new BeanInitializationException(
							"Could not find DispatcherServlet's default strategy class ["
									+ className + "] for interface [" + key
									+ "]", ex);
				} catch (LinkageError err) {
					throw new BeanInitializationException(
							"Error loading DispatcherServlet's default strategy class ["
									+ className
									+ "] for interface ["
									+ key
									+ "]: problem with class file or dependent class",
							err);
				}
			}
		} else {
			strategies = Collections.EMPTY_LIST;
		}
		return strategies;
	}
}
