/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.tqlab.sense.servlet;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.lang.StringUtils;
import org.springframework.context.ApplicationContext;
import org.springframework.context.i18n.LocaleContext;
import org.springframework.context.i18n.LocaleContextHolder;
import org.springframework.web.context.request.RequestAttributes;
import org.springframework.web.context.request.RequestContextHolder;
import org.springframework.web.context.request.ServletRequestAttributes;
import org.springframework.web.servlet.DispatcherServlet;
import org.springframework.web.servlet.HandlerAdapter;
import org.springframework.web.servlet.HandlerExceptionResolver;
import org.springframework.web.servlet.HandlerExecutionChain;
import org.springframework.web.servlet.HandlerInterceptor;
import org.springframework.web.servlet.HandlerMapping;
import org.springframework.web.servlet.LocaleResolver;
import org.springframework.web.servlet.ModelAndView;
import org.springframework.web.servlet.ModelAndViewDefiningException;
import org.springframework.web.servlet.RequestToViewNameTranslator;
import org.springframework.web.servlet.ThemeResolver;
import org.springframework.web.servlet.View;
import org.springframework.web.servlet.ViewResolver;
import org.springframework.web.util.NestedServletException;
import org.springframework.web.util.WebUtils;

import com.tqlab.sense.context.SenseServletContext;
import com.tqlab.sense.context.WebApp;

/**
 * Central dispatcher for HTTP request handlers/controllers, e.g. for web UI
 * controllers or HTTP-based remote service exporters. Dispatches to registered
 * handlers for processing a web request, providing convenient mapping and
 * exception handling facilities.
 * 
 * <p>
 * This servlet is very flexible: It can be used with just about any workflow,
 * with the installation of the appropriate adapter classes. It offers the
 * following functionality that distinguishes it from other request-driven web
 * MVC frameworks:
 * 
 * <ul>
 * <li>It is based around a JavaBeans configuration mechanism.
 * 
 * <li>It can use any {@link HandlerMapping} implementation - pre-built or
 * provided as part of an application - to control the routing of requests to
 * handler objects. Default is
 * {@link org.springframework.web.servlet.handler.BeanNameUrlHandlerMapping}, as
 * well as a
 * {@link org.springframework.web.servlet.mvc.annotation.DefaultAnnotationHandlerMapping}
 * when running on Java 5+. HandlerMapping objects can be defined as beans in
 * the servlet's application context, implementing the HandlerMapping interface,
 * overriding the default HandlerMapping if present. HandlerMappings can be
 * given any bean name (they are tested by type).
 * 
 * <li>It can use any {@link HandlerAdapter}; this allows for using any handler
 * interface. Default adapters are
 * {@link org.springframework.web.servlet.mvc.HttpRequestHandlerAdapter},
 * {@link org.springframework.web.servlet.mvc.SimpleControllerHandlerAdapter}
 * and
 * {@link org.springframework.web.servlet.mvc.throwaway.ThrowawayControllerHandlerAdapter}
 * , for Spring's {@link org.springframework.web.HttpRequestHandler},
 * {@link org.springframework.web.servlet.mvc.Controller} and
 * {@link org.springframework.web.servlet.mvc.throwaway.ThrowawayController}
 * interfaces, respectively. When running in a Java 5+ environment, a default
 * {@link org.springframework.web.servlet.mvc.annotation.AnnotationMethodHandlerAdapter}
 * will be registered as well. HandlerAdapter objects can be added as beans in
 * the application context, overriding the default HandlerAdapters. Like
 * HandlerMappings, HandlerAdapters can be given any bean name (they are tested
 * by type).
 * 
 * <li>The dispatcher's exception resolution strategy can be specified via a
 * {@link HandlerExceptionResolver}, for example mapping certain exceptions to
 * error pages. Default is none. Additional HandlerExceptionResolvers can be
 * added through the application context. HandlerExceptionResolver can be given
 * any bean name (they are tested by type).
 * 
 * <li>Its view resolution strategy can be specified via a {@link ViewResolver}
 * implementation, resolving symbolic view names into View objects. Default is
 * {@link org.springframework.web.servlet.view.InternalResourceViewResolver}.
 * ViewResolver objects can be added as beans in the application context,
 * overriding the default ViewResolver. ViewResolvers can be given any bean name
 * (they are tested by type).
 * 
 * <li>If a {@link View} or view name is not supplied by the user, then the
 * configured {@link RequestToViewNameTranslator} will translate the current
 * request into a view name. The corresponding bean name is
 * "viewNameTranslator"; the default is
 * {@link org.springframework.web.servlet.view.DefaultRequestToViewNameTranslator}.
 * 
 * <li>The dispatcher's strategy for resolving multipart requests is determined
 * by a {@link org.springframework.web.multipart.MultipartResolver}
 * implementation. Implementations for Jakarta Commons FileUpload and Jason
 * Hunter's COS are included; the typical choise is
 * {@link org.springframework.web.multipart.commons.CommonsMultipartResolver}.
 * The MultipartResolver bean name is "multipartResolver"; default is none.
 * 
 * <li>Its locale resolution strategy is determined by a {@link LocaleResolver}.
 * Out-of-the-box implementations work via HTTP accept header, cookie, or
 * session. The LocaleResolver bean name is "localeResolver"; default is
 * {@link org.springframework.web.servlet.i18n.AcceptHeaderLocaleResolver}.
 * 
 * <li>Its theme resolution strategy is determined by a {@link ThemeResolver}.
 * Implementations for a fixed theme and for cookie and session storage are
 * included. The ThemeResolver bean name is "themeResolver"; default is
 * {@link org.springframework.web.servlet.theme.FixedThemeResolver}.
 * </ul>
 * 
 * <p>
 * <b>NOTE: The <code>@RequestMapping</code> annotation will only be processed
 * if a corresponding <code>HandlerMapping</code> (for type level annotations)
 * and/or <code>HandlerAdapter</code> (for method level annotations) is present
 * in the dispatcher.</b> This is the case by default. However, if you are
 * defining custom <code>HandlerMappings</code> or <code>HandlerAdapters</code>,
 * then you need to make sure that a corresponding custom
 * <code>DefaultAnnotationHandlerMapping</code> and/or
 * <code>AnnotationMethodHandlerAdapter</code> is defined as well - provided
 * that you intend to use <code>@RequestMapping</code>.
 * 
 * <p>
 * <b>A web application can define any number of DispatcherServlets.</b> Each
 * servlet will operate in its own namespace, loading its own application
 * context with mappings, handlers, etc. Only the root application context as
 * loaded by {@link org.springframework.web.context.ContextLoaderListener}, if
 * any, will be shared.
 * 
 * @author Rod Johnson
 * @author Juergen Hoeller
 * @author Rob Harrop
 * @see org.springframework.web.HttpRequestHandler
 * @see org.springframework.web.servlet.mvc.Controller
 * @see org.springframework.web.context.ContextLoaderListener
 */
public class SenseDispatcherServlet extends DispatcherServlet {

	private static final long serialVersionUID = -757929860206741416L;

	/**
	 * Expose LocaleContext and RequestAttributes as inheritable for child
	 * threads?
	 */
	private boolean threadContextInheritable = false;

	/**
	 * Initialize the strategy objects that this servlet uses.
	 * <p>
	 * May be overridden in subclasses in order to initialize further strategy
	 * objects.
	 */
	protected void initStrategies(ApplicationContext context) {
		super.initStrategies(context);
	}

	/**
	 * Exposes the DispatcherServlet-specific request attributes and delegates
	 * to {@link #doDispatch} for the actual dispatching.
	 */
	protected void doService(HttpServletRequest request,
			HttpServletResponse response) throws Exception {
		WebApp.setSenseServletContext(new SenseServletContext());
		WebApp.setApplicationContext(this.getWebApplicationContext());

		super.doService(request, response);
		/**
		 * Clear Context.
		 */
		WebApp.setSenseServletContext(null);
		WebApp.setApplicationContext(null);
	}

	/**
	 * Process the actual dispatching to the handler.
	 * <p>
	 * The handler will be obtained by applying the servlet's HandlerMappings in
	 * order. The HandlerAdapter will be obtained by querying the servlet's
	 * installed HandlerAdapters to find the first that supports the handler
	 * class.
	 * <p>
	 * All HTTP methods are handled by this method. It's up to HandlerAdapters
	 * or handlers themselves to decide which methods are acceptable.
	 * 
	 * @param request
	 *            current HTTP request
	 * @param response
	 *            current HTTP response
	 * @throws Exception
	 *             in case of any kind of processing failure
	 */
	protected void doDispatch(HttpServletRequest request,
			HttpServletResponse response) throws Exception {
		HttpServletRequest processedRequest = request;
		HandlerExecutionChain mappedHandler = null;
		int interceptorIndex = -1;

		// Expose current LocaleResolver and request as LocaleContext.
		LocaleContext previousLocaleContext = LocaleContextHolder
				.getLocaleContext();
		LocaleContextHolder.setLocaleContext(buildLocaleContext(request),
				this.threadContextInheritable);

		// Expose current RequestAttributes to current thread.
		RequestAttributes previousRequestAttributes = RequestContextHolder
				.getRequestAttributes();
		ServletRequestAttributes requestAttributes = new ServletRequestAttributes(
				request);
		RequestContextHolder.setRequestAttributes(requestAttributes,
				this.threadContextInheritable);

		if (logger.isTraceEnabled()) {
			logger.trace("Bound request context to thread: " + request);
		}

		try {
			ModelAndView mv = null;
			boolean errorView = false;

			try {
				processedRequest = checkMultipart(request);

				// Determine handler for the current request.
				mappedHandler = getHandler(processedRequest);
				if (mappedHandler == null || mappedHandler.getHandler() == null) {
					noHandlerFound(processedRequest, response);
					return;
				}

				// Apply preHandle methods of registered interceptors.
				HandlerInterceptor[] interceptors = mappedHandler
						.getInterceptors();
				if (interceptors != null) {
					for (int i = 0; i < interceptors.length; i++) {
						HandlerInterceptor interceptor = interceptors[i];
						if (!interceptor.preHandle(processedRequest, response,
								mappedHandler.getHandler())) {
							triggerAfterCompletion(mappedHandler,
									interceptorIndex, processedRequest,
									response, null);
							return;
						}
						interceptorIndex = i;
					}
				}

				// Actually invoke the handler.
				HandlerAdapter ha = getHandlerAdapter(mappedHandler
						.getHandler());

				// 更新当前线程的SenceServletContext
				SenseServletContext ssc = WebApp
						.getCurrentSenceServletContext();
				ssc.setServletRequest(request);
				ssc.setServletResponse(response);
				String modual = null;
				String contextPath = request.getContextPath();
				if (!StringUtils.isBlank(contextPath)) {
					modual = contextPath;
					if (modual.startsWith("/")) {
						modual = modual.substring(1);
					}
				}
				ssc.setModual(modual);
				WebApp.setSenseServletContext(ssc);
				// 操作完毕

				mv = ha.handle(processedRequest, response,
						mappedHandler.getHandler());

				// Do we need view name translation?
				if (mv != null && !mv.hasView()) {
					mv.setViewName(getDefaultViewName(request));
				}

				// Apply postHandle methods of registered interceptors.
				if (interceptors != null) {
					for (int i = interceptors.length - 1; i >= 0; i--) {
						HandlerInterceptor interceptor = interceptors[i];
						interceptor.postHandle(processedRequest, response,
								mappedHandler.getHandler(), mv);
					}
				}
			} catch (ModelAndViewDefiningException ex) {
				logger.debug("ModelAndViewDefiningException encountered", ex);
				mv = ex.getModelAndView();
			} catch (Exception ex) {
				Object handler = (mappedHandler != null ? mappedHandler
						.getHandler() : null);
				mv = processHandlerException(processedRequest, response,
						handler, ex);
				errorView = (mv != null);
			}

			// Did the handler return a view to render?
			if (mv != null && !mv.wasCleared()) {
				render(mv, processedRequest, response);
				if (errorView) {
					WebUtils.clearErrorRequestAttributes(request);
				}
			} else {
				if (logger.isDebugEnabled()) {
					logger.debug("Null ModelAndView returned to DispatcherServlet with name '"
							+ getServletName()
							+ "': assuming HandlerAdapter completed request handling");
				}
			}

			// Trigger after-completion for successful outcome.
			triggerAfterCompletion(mappedHandler, interceptorIndex,
					processedRequest, response, null);
		}

		catch (Exception ex) {
			// Trigger after-completion for thrown exception.
			triggerAfterCompletion(mappedHandler, interceptorIndex,
					processedRequest, response, ex);
			throw ex;
		} catch (Error err) {
			ServletException ex = new NestedServletException(
					"Handler processing failed", err);
			// Trigger after-completion for thrown exception.
			triggerAfterCompletion(mappedHandler, interceptorIndex,
					processedRequest, response, ex);
			throw ex;
		}

		finally {
			// Clean up any resources used by a multipart request.
			if (processedRequest != request) {
				cleanupMultipart(processedRequest);
			}

			// Reset thread-bound context.
			RequestContextHolder.setRequestAttributes(
					previousRequestAttributes, this.threadContextInheritable);
			LocaleContextHolder.setLocaleContext(previousLocaleContext,
					this.threadContextInheritable);

			// Clear request attributes.
			requestAttributes.requestCompleted();
			if (logger.isTraceEnabled()) {
				logger.trace("Cleared thread-bound request context: " + request);
			}
		}
	}

	/**
	 * Trigger afterCompletion callbacks on the mapped HandlerInterceptors. Will
	 * just invoke afterCompletion for all interceptors whose preHandle
	 * invocation has successfully completed and returned true.
	 * 
	 * @param mappedHandler
	 *            the mapped HandlerExecutionChain
	 * @param interceptorIndex
	 *            index of last interceptor that successfully completed
	 * @param ex
	 *            Exception thrown on handler execution, or <code>null</code> if
	 *            none
	 * @see HandlerInterceptor#afterCompletion
	 */
	private void triggerAfterCompletion(HandlerExecutionChain mappedHandler,
			int interceptorIndex, HttpServletRequest request,
			HttpServletResponse response, Exception ex) throws Exception {

		// Apply afterCompletion methods of registered interceptors.
		if (mappedHandler != null) {
			HandlerInterceptor[] interceptors = mappedHandler.getInterceptors();
			if (interceptors != null) {
				for (int i = interceptorIndex; i >= 0; i--) {
					HandlerInterceptor interceptor = interceptors[i];
					try {
						interceptor.afterCompletion(request, response,
								mappedHandler.getHandler(), ex);
					} catch (Throwable ex2) {
						logger.error(
								"HandlerInterceptor.afterCompletion threw exception",
								ex2);
					}
				}
			}
		}
	}

}
