/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.tqlab.sense.servlet.handler;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;

import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.core.annotation.AnnotationUtils;
import org.springframework.stereotype.Controller;
import org.springframework.util.Assert;
import org.springframework.util.ObjectUtils;
import org.springframework.util.ReflectionUtils;
import org.springframework.util.StringUtils;
import org.springframework.web.HttpRequestMethodNotSupportedException;
import org.springframework.web.bind.ServletRequestBindingException;
import org.springframework.web.bind.UnsatisfiedServletRequestParameterException;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.servlet.handler.AbstractDetectingUrlHandlerMapping;

import com.tqlab.sense.tiles.TileParameters;
import com.tqlab.sense.tiles.TileTool;

/**
 * @author John Lee
 * 
 */
public class OSGIAnnotationHandlerMapping extends
		AbstractDetectingUrlHandlerMapping implements OSGIBundleHandler {

	private boolean useDefaultSuffixPattern = true;

	private final Map<Class<?>, RequestMapping> cachedMappings = new HashMap<Class<?>, RequestMapping>();

	/**
	 * Set whether to register paths using the default suffix pattern as well:
	 * i.e. whether "/users" should be registered as "/users.*" and "/users/"
	 * too.
	 * <p>
	 * Default is "true". Turn this convention off if you intend to interpret
	 * your <code>@RequestMapping</code> paths strictly.
	 * <p>
	 * Note that paths which include a ".xxx" suffix or end with "/" already
	 * will not be transformed using the default suffix pattern in any case.
	 */
	public void setUseDefaultSuffixPattern(boolean useDefaultSuffixPattern) {
		this.useDefaultSuffixPattern = useDefaultSuffixPattern;
	}

	/**
	 * Checks for presence of the
	 * {@link org.springframework.web.bind.annotation.RequestMapping} annotation
	 * on the handler class and on any of its methods.
	 */
	@Override
	protected String[] determineUrlsForHandler(String beanName) {
		ApplicationContext context = getApplicationContext();
		return determineUrlsForHandler(beanName, context);
	}

	protected String[] determineUrlsForHandler(String beanName,
			ApplicationContext context) {
		Class<?> handlerType = context.getType(beanName);
		RequestMapping mapping = context.findAnnotationOnBean(beanName,
				RequestMapping.class);
		if (mapping != null) {
			// @RequestMapping found at type level
			this.cachedMappings.put(handlerType, mapping);
			Set<String> urls = new LinkedHashSet<String>();
			String[] typeLevelPatterns = mapping.value();
			if (typeLevelPatterns.length > 0) {
				// @RequestMapping specifies paths at type level
				String[] methodLevelPatterns = determineUrlsForHandlerMethods(
						handlerType, true);
				for (String typeLevelPattern : typeLevelPatterns) {
					if (!typeLevelPattern.startsWith("/")) {
						typeLevelPattern = "/" + typeLevelPattern;
					}
					boolean hasEmptyMethodLevelMappings = false;
					for (String methodLevelPattern : methodLevelPatterns) {
						if (methodLevelPattern == null) {
							hasEmptyMethodLevelMappings = true;
						} else {
							String combinedPattern = getPathMatcher().combine(
									typeLevelPattern, methodLevelPattern);
							addUrlsForPath(urls, combinedPattern);
						}
					}
					if (hasEmptyMethodLevelMappings
							|| org.springframework.web.servlet.mvc.Controller.class
									.isAssignableFrom(handlerType)) {
						addUrlsForPath(urls, typeLevelPattern);
					}
				}
				return StringUtils.toStringArray(urls);
			} else {
				// actual paths specified by @RequestMapping at method level
				return determineUrlsForHandlerMethods(handlerType, false);
			}
		} else if (AnnotationUtils
				.findAnnotation(handlerType, Controller.class) != null) {
			// @RequestMapping to be introspected at method level
			return determineUrlsForHandlerMethods(handlerType, false);
		} else {
			return null;
		}
	}

	/**
	 * Derive URL mappings from the handler's method-level mappings.
	 * 
	 * @param handlerType
	 *            the handler type to introspect
	 * @param hasTypeLevelMapping
	 *            whether the method-level mappings are nested within a
	 *            type-level mapping
	 * @return the array of mapped URLs
	 */
	protected String[] determineUrlsForHandlerMethods(Class<?> handlerType,
			final boolean hasTypeLevelMapping) {
		String[] subclassResult = determineUrlsForHandlerMethods(handlerType);
		if (subclassResult != null) {
			return subclassResult;
		}

		final Set<String> urls = new LinkedHashSet<String>();
		Set<Class<?>> handlerTypes = new LinkedHashSet<Class<?>>();
		handlerTypes.add(handlerType);
		handlerTypes.addAll(Arrays.asList(handlerType.getInterfaces()));
		for (Class<?> currentHandlerType : handlerTypes) {
			ReflectionUtils.doWithMethods(currentHandlerType,
					new ReflectionUtils.MethodCallback() {
						public void doWith(Method method) {
							RequestMapping mapping = AnnotationUtils
									.findAnnotation(method,
											RequestMapping.class);
							if (mapping != null) {
								String[] mappedPatterns = mapping.value();
								if (mappedPatterns.length > 0) {
									for (String mappedPattern : mappedPatterns) {
										if (!hasTypeLevelMapping
												&& !mappedPattern
														.startsWith("/")) {
											mappedPattern = "/" + mappedPattern;
										}
										addUrlsForPath(urls, mappedPattern);
									}
								} else if (hasTypeLevelMapping) {
									// empty method-level RequestMapping
									urls.add(null);
								}
							}
						}
					}, ReflectionUtils.USER_DECLARED_METHODS);
		}
		return StringUtils.toStringArray(urls);
	}

	/**
	 * Derive URL mappings from the handler's method-level mappings.
	 * 
	 * @param handlerType
	 *            the handler type to introspect
	 * @return the array of mapped URLs
	 */
	protected String[] determineUrlsForHandlerMethods(Class<?> handlerType) {
		return null;
	}

	/**
	 * Add URLs and/or URL patterns for the given path.
	 * 
	 * @param urls
	 *            the Set of URLs for the current bean
	 * @param path
	 *            the currently introspected path
	 */
	protected void addUrlsForPath(Set<String> urls, String path) {
		urls.add(path);
		if (this.useDefaultSuffixPattern && path.indexOf('.') == -1
				&& !path.endsWith("/")) {
			urls.add(path + ".*");
			urls.add(path + "/");
		}
	}

	/**
	 * Validate the given annotated handler against the current request.
	 * 
	 * @see #validateMapping
	 */
	@Override
	protected void validateHandler(Object handler, HttpServletRequest request)
			throws Exception {
		RequestMapping mapping = this.cachedMappings.get(handler.getClass());
		if (mapping == null) {
			mapping = AnnotationUtils.findAnnotation(handler.getClass(),
					RequestMapping.class);
		}
		if (mapping != null) {
			validateMapping(mapping, request);
		}
	}

	/**
	 * Validate the given type-level mapping metadata against the current
	 * request, checking HTTP request method and parameter conditions.
	 * 
	 * @param mapping
	 *            the mapping metadata to validate
	 * @param request
	 *            current HTTP request
	 * @throws Exception
	 *             if validation failed
	 */
	protected void validateMapping(RequestMapping mapping,
			HttpServletRequest request) throws Exception {
		RequestMethod[] mappedMethods = mapping.method();
		if (!ServletAnnotationMappingUtils.checkRequestMethod(mappedMethods,
				request)) {
			String[] supportedMethods = new String[mappedMethods.length];
			for (int i = 0; i < mappedMethods.length; i++) {
				supportedMethods[i] = mappedMethods[i].name();
			}
			throw new HttpRequestMethodNotSupportedException(
					request.getMethod(), supportedMethods);
		}

		String[] mappedParams = mapping.params();
		if (!ServletAnnotationMappingUtils.checkParameters(mappedParams,
				request)) {
			throw new UnsatisfiedServletRequestParameterException(mappedParams,
					request.getParameterMap());
		}

		String[] mappedHeaders = mapping.headers();
		if (!ServletAnnotationMappingUtils.checkHeaders(mappedHeaders, request)) {
			throw new ServletRequestBindingException("Header conditions \""
					+ StringUtils.arrayToDelimitedString(mappedHeaders, ", ")
					+ "\" not met for actual request");
		}
	}

	public void onStartBundle(ApplicationContext applicationContext,
			String bundleName) {
		if (logger.isDebugEnabled()) {
			logger.debug("Looking for URL mappings in application context: "
					+ applicationContext);
		}
		String[] beanNames = applicationContext
				.getBeanNamesForType(Object.class);

		// Take any bean name that we can determine URLs for.
		for (String beanName : beanNames) {
			String[] urls = determineUrlsForHandler(beanName,
					applicationContext);
			if (!ObjectUtils.isEmpty(urls)) {
				for (String urlPath : urls) {
					registerHandler(urlPath, beanName, applicationContext);
				}
			} else {
				if (logger.isDebugEnabled()) {
					logger.debug("Rejected bean name '" + beanName
							+ "': no URL paths identified");
				}
			}
		}
	}

	public void onStopBundle(ApplicationContext applicationContext,
			String bundleName) {
		String[] beanNames = applicationContext
				.getBeanNamesForType(Object.class);
		// Take any bean name that we can determine URLs for.
		for (String beanName : beanNames) {
			String[] urls = determineUrlsForHandler(beanName,
					applicationContext);
			if (!ObjectUtils.isEmpty(urls)) {
				for (String urlPath : urls) {
					@SuppressWarnings("unchecked")
					Map<String, Object> handlerMap = (Map<String, Object>) getFieldValue(
							this, "handlerMap");
					handlerMap.remove(urlPath);
				}
			} else {
				if (logger.isDebugEnabled()) {
					logger.debug("Rejected bean name '" + beanName
							+ "': no URL paths identified");
				}
			}
		}

	}

	protected void registerHandler(String urlPath, Object handler)
			throws BeansException, IllegalStateException {
		super.registerHandler(urlPath, handler);
	}

	protected void registerHandler(String urlPath, Object handler,
			ApplicationContext applicationContext) throws BeansException,
			IllegalStateException {
		Assert.notNull(urlPath, "URL path must not be null");
		Assert.notNull(handler, "Handler object must not be null");
		Object resolvedHandler = handler;

		Boolean lazyInitHandlers = (Boolean) getFieldValue(this,
				"lazyInitHandlers");
		lazyInitHandlers = (lazyInitHandlers == null ? false : lazyInitHandlers);

		// Eagerly resolve handler if referencing singleton via name.
		if (!lazyInitHandlers && handler instanceof String) {
			String handlerName = (String) handler;
			if (applicationContext.isSingleton(handlerName)) {
				resolvedHandler = applicationContext.getBean(handlerName);
			}
		}

		Object mappedHandler = this.getHandlerMap().get(urlPath);
		if (mappedHandler != null) {
			if (mappedHandler != resolvedHandler) {
				throw new IllegalStateException("Cannot map "
						+ getHandlerDescription(handler) + " to URL path ["
						+ urlPath + "]: There is already "
						+ getHandlerDescription(mappedHandler) + " mapped.");
			}
		} else {
			if (urlPath.equals("/")) {
				if (logger.isInfoEnabled()) {
					logger.info("Root mapping to "
							+ getHandlerDescription(handler));
				}
				setRootHandler(resolvedHandler);
			} else if (urlPath.equals("/*")) {
				if (logger.isInfoEnabled()) {
					logger.info("Default mapping to "
							+ getHandlerDescription(handler));
				}
				setDefaultHandler(resolvedHandler);
			} else {
				@SuppressWarnings("unchecked")
				Map<String, Object> handlerMap = (Map<String, Object>) getFieldValue(
						this, "handlerMap");
				handlerMap.put(urlPath, resolvedHandler);
				if (logger.isInfoEnabled()) {
					logger.info("Mapped URL path [" + urlPath + "] onto "
							+ getHandlerDescription(handler));
				}
			}
		}
	}

	private String getHandlerDescription(Object handler) {
		return "handler "
				+ (handler instanceof String ? "'" + handler + "'"
						: "of type [" + handler.getClass() + "]");
	}

	@Override
	protected Object getHandlerInternal(HttpServletRequest request)
			throws Exception {

		String lookupPath = this.getLookupPathForRequest(request);
		if (null == lookupPath) {
			return super.getHandlerInternal(request);
		}

		Object handler = lookupHandler(lookupPath, request);
		if (handler == null) {
			// We need to care for the default handler directly, since we need
			// to
			// expose the PATH_WITHIN_HANDLER_MAPPING_ATTRIBUTE for it as well.
			Object rawHandler = null;
			if ("/".equals(lookupPath)) {
				rawHandler = getRootHandler();
			}
			if (rawHandler == null) {
				rawHandler = getDefaultHandler();
			}
			if (rawHandler != null) {
				validateHandler(rawHandler, request);
				handler = buildPathExposingHandler(rawHandler, lookupPath,
						lookupPath, null);
			}
		}
		if (handler != null && logger.isDebugEnabled()) {
			logger.debug("Mapping [" + lookupPath + "] to handler '" + handler
					+ "'");
		} else if (handler == null && logger.isTraceEnabled()) {
			logger.trace("No handler mapping found for [" + lookupPath + "]");
		}
		return handler;
	}

	/**
	 * 
	 * @param request
	 * @return
	 */
	@SuppressWarnings("unchecked")
	private String getLookupPathForRequest(HttpServletRequest request) {
		TileParameters<String, Object> params = (TileParameters<String, Object>) request
				.getAttribute(TileTool.TILE_LIST_HANLER);
		if (null == params) {
			return null;
		}
		return "/" + params.getModual() + ":" + params.getTemplate();
	}

	private static Object getFieldValue(Object instance, String fieldName) {
		Object obj = null;
		List<Class<?>> list = new ArrayList<Class<?>>();
		list.add(instance.getClass());
		Class<?> superClass = instance.getClass().getSuperclass();
		while (superClass != Object.class) {
			list.add(superClass);
			superClass = superClass.getSuperclass();
		}

		for (Class<?> clazz : list) {
			try {
				Field field = clazz.getDeclaredField(fieldName);
				field.setAccessible(true);
				obj = field.get(instance);
				break;
			} catch (Exception e) {

			}
		}

		return obj;

	}
}
