/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.tqlab.sense.servlet.view.velocity;

import java.io.StringWriter;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.velocity.Template;
import org.apache.velocity.context.Context;
import org.apache.velocity.exception.ResourceNotFoundException;
import org.apache.velocity.runtime.RuntimeConstants;
import org.springframework.beans.factory.BeanFactoryUtils;
import org.springframework.beans.factory.NoSuchBeanDefinitionException;
import org.springframework.context.ApplicationContextException;
import org.springframework.web.servlet.view.velocity.VelocityToolboxView;

import com.tqlab.sense.servlet.view.velocity.tools.ToolBox;
import com.tqlab.sense.tiles.TileTool;

/**
 * 
 * @author John Lee
 * 
 */
public class SenseVelocityLayoutView extends VelocityToolboxView {

	public static final String DEFAULT_SCREEN_CONTENT_KEY = "screen_content";
	public static final String DEFAULT_COMPONENT_CONTENT_KEY = "_component_placeholder_";
	/**
	 * The default {@link #setLayoutUrl(String) layout url}.
	 */
	public static final String DEFAULT_LAYOUT_URL = "default.vm";

	/**
	 * The default {@link #setLayoutKey(String) layout key}.
	 */
	public static final String DEFAULT_LAYOUT_KEY = "layout";

	private String layoutUrl = DEFAULT_LAYOUT_URL;
	private String layoutKey = DEFAULT_LAYOUT_KEY;

	/** 当前web request返回的view的vm url */
	public static final String URL_UNDER_RENDERING_KEY = "rendering_url";
	public static final String SCREEN_CONTEXT_KEY = "screen_context_key";

	private String screenContentKey = DEFAULT_SCREEN_CONTENT_KEY;

	private boolean isLayout = true;

	/**
	 * 
	 * @param velocityContext
	 * @param request
	 * @throws Exception
	 */
	private void exposeToolsAttributes(Context velocityContext,
			HttpServletRequest request) throws Exception {

		// Title tool
		TileTool tileTool = (TileTool) getApplicationContext().getBean(
				"tileTool");
		velocityContext.put("tile", tileTool);

		try {
			// 获取AlipayToolBox对像
			ToolBox toolBox = (ToolBox) BeanFactoryUtils
					.beanOfTypeIncludingAncestors(getApplicationContext(),
							ToolBox.class, true, false);
			Map<String, Object> map = toolBox.getToolsContext(request);
			for (Object key : map.keySet()) {
				String keyStr = key.toString();
				velocityContext.put(keyStr, map.get(keyStr));
			}
		} catch (NoSuchBeanDefinitionException e) {
			throw new ApplicationContextException(
					"Must define a single ToolBox bean in this web application context "
							+ "(may be inherited): ToolBox is the usual implementation. "
							+ "This bean may be given any name.", e);

		}

	}

	/**
	 * Set the layout template to use. Default is {@link #DEFAULT_LAYOUT_URL
	 * "layout.vm"}.
	 * 
	 * @param layoutUrl
	 *            the template location (relative to the template root
	 *            directory)
	 */
	public void setLayoutUrl(String layoutUrl) {
		this.layoutUrl = layoutUrl;
	}

	/**
	 * Set the context key used to specify an alternate layout to be used
	 * instead of the default layout. Screen content templates can override the
	 * layout template that they wish to be wrapped with by setting this value
	 * in the template, for example:<br>
	 * <code>#set( $layout = "MyLayout.vm" )</code>
	 * <p>
	 * Default key is {@link #DEFAULT_LAYOUT_KEY "layout"}, as illustrated
	 * above.
	 * 
	 * @param layoutKey
	 *            the name of the key you wish to use in your screen content
	 *            templates to override the layout template
	 */
	public void setLayoutKey(String layoutKey) {
		this.layoutKey = layoutKey;
	}

	/**
	 * Set the name of the context key that will hold the content of the screen
	 * within the layout template. This key must be present in the layout
	 * template for the current screen to be rendered.
	 * <p>
	 * Default is {@link #DEFAULT_SCREEN_CONTENT_KEY "screen_content"}: accessed
	 * in VTL as <code>$screen_content</code>.
	 * 
	 * @param screenContentKey
	 *            the name of the screen content key to use
	 */
	public void setScreenContentKey(String screenContentKey) {
		this.screenContentKey = screenContentKey;
	}

	/**
	 * Overrides <code>VelocityView.checkTemplate()</code> to additionally check
	 * that both the layout template and the screen content template can be
	 * loaded. Note that during rendering of the screen content, the layout
	 * template can be changed which may invalidate any early checking done
	 * here.
	 */
	protected void checkTemplate() throws ApplicationContextException {
		// super.checkTemplate();

		try {
			// Check that we can get the template, even if we might subsequently
			// get it again.
			if (this.isLayout) {
				getTemplate(this.layoutUrl);
			}
		} catch (ResourceNotFoundException ex) {
			throw new ApplicationContextException(
					"Cannot find Velocity template for URL ["
							+ this.layoutUrl
							+ "]: Did you specify the correct resource loader path?",
					ex);
		} catch (Exception ex) {
			throw new ApplicationContextException(
					"Could not load Velocity template for URL ["
							+ this.layoutUrl + "]", ex);
		}
	}

	/**
	 * Process the model map by merging it with the Velocity template. Output is
	 * directed to the servlet response.
	 * <p>
	 * This method can be overridden if custom behavior is needed.
	 */
	@SuppressWarnings("unchecked")
	protected void renderMergedTemplateModel(
			@SuppressWarnings("rawtypes") final Map model,
			HttpServletRequest request, HttpServletResponse response)
			throws Exception {

		exposeHelpers(model, request);
		Context velocityContext = createVelocityContext(model, request,
				response);
		exposeHelpers(velocityContext, request, response);
		exposeToolAttributes(velocityContext, request);
		// exposeTileToolAttributes(velocityContext);
		exposeToolsAttributes(velocityContext, request);
		try {
			if (this.isLayout) {
				request.setAttribute(SCREEN_CONTEXT_KEY, model);
				request.setAttribute(URL_UNDER_RENDERING_KEY, getUrl());
				doRender(velocityContext, response);
			} else {
				// 对于Component 类型的模板渲染暂时用这种方式实现。
				doRender(velocityContext, request, response);
			}
		} finally {
		}
	}

	/**
	 * Overrides the normal rendering process in order to pre-process the
	 * Context, merging it with the screen template into a single value
	 * (identified by the value of screenContentKey). The layout template is
	 * then merged with the modified Context in the super class.
	 */
	protected void doRender(Context context, HttpServletResponse response)
			throws Exception {

		renderScreenContent(context);

		// Velocity context now includes any mappings that were defined
		// (via #set) in screen content template.
		// The screen template can overrule the layout by doing
		// #set( $layout = "MyLayout.vm" )
		String layoutUrlToUse = (String) context.get(this.layoutKey);
		if (layoutUrlToUse != null) {
			if (logger.isDebugEnabled()) {
				logger.debug("Screen content template has requested layout ["
						+ layoutUrlToUse + "]");
			}
		} else {
			// No explicit layout URL given -> use default layout of this view.
			layoutUrlToUse = this.layoutUrl;
		}

		mergeTemplate(getTemplate(layoutUrlToUse), context, response);

	}

	protected void doRender(Context context, HttpServletRequest request,
			HttpServletResponse response) throws Exception {

		if (logger.isDebugEnabled()) {
			logger.debug("Rendering component content template [" + getUrl()
					+ "]");
		}

		StringWriter sw = new StringWriter();
		sw.append("");
		Template componentContentTemplate = getTemplate(getUrl());
		componentContentTemplate.merge(context, sw);

		// Put rendered content into Velocity context.
		request.setAttribute(DEFAULT_COMPONENT_CONTENT_KEY, sw.toString());

	}

	/**
	 * The resulting context contains any mappings from render, plus screen
	 * content.
	 */
	private void renderScreenContent(Context velocityContext) throws Exception {
		if (logger.isDebugEnabled()) {
			logger.debug("Rendering screen content template [" + getUrl() + "]");
		}

		StringWriter sw = new StringWriter();
		Template screenContentTemplate = getTemplate(getUrl());
		screenContentTemplate.merge(velocityContext, sw);

		// Put rendered content into Velocity context.
		velocityContext.put(this.screenContentKey, sw.toString());
	}

	protected void applyContentType(HttpServletResponse response) {
		super.applyContentType(response);
		String encoding = (String) getVelocityEngine().getProperty(
				RuntimeConstants.OUTPUT_ENCODING);
		response.setCharacterEncoding(encoding);
	}

	/**
	 * @return Returns the isLayout.
	 */
	public boolean isLayout() {
		return isLayout;
	}

	/**
	 * @param isLayout
	 *            The isLayout to set.
	 */
	public void setLayout(boolean isLayout) {
		this.isLayout = isLayout;
	}
}
