/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.tqlab.sense.tiles;

import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Stack;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.log4j.Logger;
import org.springframework.util.Assert;
import org.springframework.web.servlet.HandlerAdapter;
import org.springframework.web.servlet.HandlerExecutionChain;
import org.springframework.web.servlet.ModelAndView;
import org.springframework.web.servlet.ModelAndViewDefiningException;
import org.springframework.web.servlet.View;
import org.springframework.web.servlet.ViewResolver;

import com.tqlab.sense.context.WebApp;
import com.tqlab.sense.servlet.handler.OSGIAnnotationHandlerMapping;
import com.tqlab.sense.servlet.handler.OSGIAnnotationMethodHandlerAdapter;

/**
 * 
 * @author John Lee
 * 
 */
public class TileTool implements Renderable {

	public static final String TILE_NAME_SEPARATOR = ":";
	public static final String DEFAULT_MODUAL = "home";
	public static final String TILE_LIST_HANLER = "_tile_list_handler";

	private String ATRRIBUTE_PATH = "_attr_path_";
	private String DEFAULT_COMPONENT_CONTENT_KEY = "_component_placeholder_";

	private Stack<TileParameters<String, Object>> tileParameterStack = new Stack<TileParameters<String, Object>>();

	private Logger logger = Logger.getLogger(getClass());

	/**
	 * 设置tile的模板。
	 * 
	 * @param template
	 *            tile模板名
	 * @return
	 */
	public TileTool setTemplate(String template) {
		Assert.hasText(template);
		String modual = WebApp.getCurrentSenceServletContext().getModual();

		int pos = template.indexOf(TILE_NAME_SEPARATOR);
		if (pos > 0) {
			modual = template.substring(0, pos);
			template = template.substring(pos + 1);
		}

		TileParameters<String, Object> params = getTileParameters();
		params.setModual(modual);
		params.setTemplate(template);
		return this;
	}

	/**
	 * 设置tile的参数。
	 * 
	 * <p>
	 * 这些参数将被保存在一个一次性的<code>Map</code>中，当render成功以后，该map就被丢弃，以便再次调用该tile。
	 * </p>
	 * 
	 * @param name
	 *            属性名
	 * @param value
	 *            对象
	 * 
	 * @return <code>TileTool</code>本身，以方便模板中的操作
	 */
	@SuppressWarnings("unchecked")
	public TileTool setParameter(String name, Object value) {
		@SuppressWarnings("rawtypes")
		TileParameters params = getTileParameters();
		params.put(name, value);
		return this;
	}

	public String render() {
		HttpServletRequest httpRequest = WebApp.getHttpServletRequest();
		HttpServletResponse httpResponse = WebApp.getHttpServletResponse();
		TileParameters<String, Object> params = getTileParameters();
		return render(httpRequest, httpResponse, params);
	}

	@SuppressWarnings("rawtypes")
	protected String render(HttpServletRequest request,
			HttpServletResponse response, TileParameters<String, Object> params) {
		Assert.hasText(params.getTemplate(), "Tile did not set template");

		HandlerExecutionChain mappedHandler = null;

		try {
			ModelAndView mv = null;
			try {
				List handlerMappings = WebApp.getHandlerMappings();
				List handlerAdapters = WebApp.getHandlerAdapters();

				if (handlerMappings == null) {
					return "handlerMappings is null";
				} else if (handlerAdapters == null) {
					return "handlerAdapters is null";
				}

				OSGIAnnotationHandlerMapping tile = null;
				HandlerAdapter adapter = null;

				for (Iterator i = handlerMappings.iterator(); i.hasNext();) {
					Object o = i.next();
					if (o instanceof OSGIAnnotationHandlerMapping) {
						tile = (OSGIAnnotationHandlerMapping) o;
						break;
					}
				}

				for (Iterator i = handlerAdapters.iterator(); i.hasNext();) {
					Object o = i.next();
					if (o instanceof OSGIAnnotationMethodHandlerAdapter) {
						adapter = (OSGIAnnotationMethodHandlerAdapter) o;
						break;
					}
				}

				if (tile == null) {
					return "OSGIAnnotationHandlerMapping is null";
				}

				if (adapter == null) {
					return "OSGIAnnotationMethodHandlerAdapter is null";
				}

				request.setAttribute(TILE_LIST_HANLER, params);
				mappedHandler = tile.getHandler(request);

				if (mappedHandler != null) {
					mv = adapter.handle(request, response,
							mappedHandler.getHandler());
				} else {
					// 创建默认的ModelAndView
					mv = new ModelAndView(params.getTemplate())
							.addAllObjects(params);
				}

			} catch (ModelAndViewDefiningException ex) {
				logger.debug("ModelAndViewDefiningException encountered", ex);
				mv = ex.getModelAndView();
			} catch (Exception ex) {
				ex.printStackTrace();
			}

			render(mv, request, response);

			String str = request.getAttribute(DEFAULT_COMPONENT_CONTENT_KEY)
					.toString();
			return str;
		} catch (Exception ex) {
			ex.printStackTrace();
		} finally {
			if (logger.isDebugEnabled()) {
				logger.debug("Cleared thread-bound request context: " + request);
			}
			request.removeAttribute(ATRRIBUTE_PATH);
			request.removeAttribute(DEFAULT_COMPONENT_CONTENT_KEY);
		}

		return null;
	}

	@SuppressWarnings({ "rawtypes", "unchecked" })
	protected void render(ModelAndView mv, HttpServletRequest request,
			HttpServletResponse response) throws Exception {

		// Did the handler return a view to render?
		if (mv == null || mv.wasCleared()) {
			if (logger.isDebugEnabled()) {
				logger.debug("Null ModelAndView returned to Component with name '"
						+ "': assuming HandlerAdapter completed request handling");
			}
		}

		try {
			View view = null;

			if (mv.isReference()) {
				// We need to resolve the view name.
				view = resolveViewName("tile:" + mv.getViewName(),
						mv.getModel(), response.getLocale(), request);
				if (view == null) {
					throw new ServletException(
							"Could not resolve view with name '"
									+ mv.getViewName().toString() + "'");
				}
			} else {
				// No need to lookup: the ModelAndView object contains the
				// actual View object.
				view = mv.getView();
				if (view == null) {
					throw new ServletException("ModelAndView [" + mv
							+ "] neither contains a view name nor a "
							+ "View object'");
				}
			}
			// Delegate to the View object for rendering.
			if (logger.isDebugEnabled()) {
				logger.debug("Rendering view [" + view + "]");
			}

			Object params = request.getAttribute(TILE_LIST_HANLER);
			if (params != null) {
				mv.addAllObjects((Map) params);
			}

			view.render(mv.getModel(), request, response);
		} finally {

		}
	}

	@SuppressWarnings("rawtypes")
	protected View resolveViewName(String viewName, Map model, Locale locale,
			HttpServletRequest request) throws Exception {
		try {
			ViewResolver viewResolver = (ViewResolver) WebApp
					.getApplicationContext().getBean("velocityViewResolver");
			View view = viewResolver.resolveViewName(viewName, locale);
			if (view != null) {
				return view;
			}
			return null;
		} finally {
		}
	}

	/**
	 * 取得栈顶的参数。
	 */
	protected TileParameters<String, Object> getTileParameters() {
		if (tileParameterStack.isEmpty()) {
			tileParameterStack.push(new TileParameters<String, Object>());
		}

		return tileParameterStack.peek();
	}

	public String toString() {
		return render();
	}
}
